@@ -1,7 +1,10 @@
 package com.fasterxml.jackson.core.sym;
 
 import java.util.Arrays;
+import java.util.BitSet;
+import java.util.concurrent.atomic.AtomicReference;
 
+import com.fasterxml.jackson.core.JsonFactory;
 import com.fasterxml.jackson.core.util.InternCache;
 
 /**
@@ -19,7 +22,7 @@ import com.fasterxml.jackson.core.util.InternCache;
  * maps/sets, that hash codes are uniformly distributed. Also, collisions
  * are slightly more expensive than with HashMap or HashSet, since hash codes
  * are not used in resolving collisions; that is, equals() comparison is
- * done with all symbols in same bucket index.<br />
+ * done with all symbols in same bucket index.<br>
  * Finally, rehashing is also more expensive, as hash codes are not
  * stored; rehashing requires all entries' hash codes to be recalculated.
  * Reason for not storing hash codes is reduced memory usage, hoping
@@ -43,31 +46,34 @@ public final class CharsToNameCanonicalizer
 {
     /* If we use "multiply-add" based hash algorithm, this is the multiplier
      * we use.
+     *<p>
+     * Note that JDK uses 31; but it seems that 33 produces fewer collisions,
+     * at least with tests we have.
      */
     public final static int HASH_MULT = 33;
-    
+
     /**
      * Default initial table size. Shouldn't be miniscule (as there's
      * cost to both array realloc and rehashing), but let's keep
-     * it reasonably small nonetheless. For systems that properly 
+     * it reasonably small. For systems that properly 
      * reuse factories it doesn't matter either way; but when
      * recreating factories often, initial overhead may dominate.
      */
-    protected static final int DEFAULT_TABLE_SIZE = 64;
+    private static final int DEFAULT_T_SIZE = 64;
 
     /**
      * Let's not expand symbol tables past some maximum size;
      * this should protected against OOMEs caused by large documents
-     * with uniquer (~= random) names.
+     * with unique (~= random) names.
      */
-    protected static final int MAX_TABLE_SIZE = 0x10000; // 64k entries == 256k mem
+    private static final int MAX_T_SIZE = 0x10000; // 64k entries == 256k mem
 
     /**
      * Let's only share reasonably sized symbol tables. Max size set to 3/4 of 16k;
      * this corresponds to 64k main hash index. This should allow for enough distinct
      * names for almost any case.
      */
-    final static int MAX_ENTRIES_FOR_REUSE = 12000;
+    static final int MAX_ENTRIES_FOR_REUSE = 12000;
 
     /**
      * Also: to thwart attacks based on hash collisions (which may or may not
@@ -78,24 +84,12 @@ public final class CharsToNameCanonicalizer
      * Note: longest chain we have been able to produce without malicious
      * intent has been 38 (with "com.fasterxml.jackson.core.main.TestWithTonsaSymbols");
      * our setting should be reasonable here.
+     *<p>
+     * Also note that value was lowered from 255 (2.3 and earlier) to 100 for 2.4
      * 
      * @since 2.1
      */
-    final static int MAX_COLL_CHAIN_LENGTH = 255;
-
-    /**
-     * And to support reduce likelihood of accidental collisons causing
-     * exceptions, let's prevent reuse of tables with long collision
-     * overflow lists as well.
-     * 
-     * @since 2.1
-     */
-    final static int MAX_COLL_CHAIN_FOR_REUSE  = 63;
-    
-    final static CharsToNameCanonicalizer sBootstrapSymbolTable;
-    static {
-        sBootstrapSymbolTable = new CharsToNameCanonicalizer();
-    }
+    static final int MAX_COLL_CHAIN_LENGTH = 100;
 
     /*
     /**********************************************************
@@ -109,7 +103,15 @@ public final class CharsToNameCanonicalizer
      * defined, and child instance is released (call to <code>release</code>),
      * parent's shared tables may be updated from the child instance.
      */
-    protected CharsToNameCanonicalizer _parent;
+    final private CharsToNameCanonicalizer _parent;
+
+    /**
+     * Member that is only used by the root table instance: root
+     * passes immutable state info child instances, and children
+     * may return new state if they add entries to the table.
+     * Child tables do NOT use the reference.
+     */
+    final private AtomicReference<TableInfo> _tableInfo;
 
     /**
      * Seed value we use as the base to make hash codes non-static between
@@ -120,20 +122,18 @@ public final class CharsToNameCanonicalizer
      * 
      * @since 2.1
      */
-    final private int _hashSeed;
-    
-    /**
-     * Whether canonical symbol Strings are to be intern()ed before added
-     * to the table or not
-     */
-    final protected boolean _intern;
+    final private int _seed;
+
+    final private int _flags;
 
     /**
      * Whether any canonicalization should be attempted (whether using
-     * intern or not)
+     * intern or not.
+     *<p>
+     * NOTE: non-final since we may need to disable this with overflow.
      */
-    final protected boolean _canonicalize;
-    
+    private boolean _canonicalize;
+
     /*
     /**********************************************************
     /* Actual symbol table data
@@ -144,7 +144,7 @@ public final class CharsToNameCanonicalizer
      * Primary matching symbols; it's expected most match occur from
      * here.
      */
-    protected String[] _symbols;
+    private String[] _symbols;
 
     /**
      * Overflow buckets; if primary doesn't match, lookup is done
@@ -153,27 +153,27 @@ public final class CharsToNameCanonicalizer
      * Note: Number of buckets is half of number of symbol entries, on
      * assumption there's less need for buckets.
      */
-    protected Bucket[] _buckets;
+    private Bucket[] _buckets;
 
     /**
      * Current size (number of entries); needed to know if and when
      * rehash.
      */
-    protected int _size;
+    private int _size;
 
     /**
      * Limit that indicates maximum size this instance can hold before
      * it needs to be expanded and rehashed. Calculated using fill
      * factor passed in to constructor.
      */
-    protected int _sizeThreshold;
+    private int _sizeThreshold;
 
     /**
      * Mask used to get index from hash values; equal to
      * <code>_buckets.length - 1</code>, when _buckets.length is
      * a power of two.
      */
-    protected int _indexMask;
+    private int _indexMask;
 
     /**
      * We need to keep track of the longest collision list; this is needed
@@ -182,8 +182,8 @@ public final class CharsToNameCanonicalizer
      * 
      * @since 2.1
      */
-    protected int _longestCollisionList;
-    
+    private int _longestCollisionList;
+
     /*
     /**********************************************************
     /* State regarding shared arrays
@@ -191,16 +191,93 @@ public final class CharsToNameCanonicalizer
      */
 
     /**
-     * Flag that indicates if any changes have been made to the data;
-     * used to both determine if bucket array needs to be copied when
-     * (first) change is made, and potentially if updated bucket list
-     * is to be resync'ed back to master instance.
+     * Flag that indicates whether underlying data structures for
+     * the main hash area are shared or not. If they are, then they
+     * need to be handled in copy-on-write way, i.e. if they need
+     * to be modified, a copy needs to be made first; at this point
+     * it will not be shared any more, and can be modified.
+     *<p>
+     * This flag needs to be checked both when adding new main entries,
+     * and when adding new collision list queues (i.e. creating a new
+     * collision list head entry)
      */
-    protected boolean _dirty;
+    private boolean _hashShared;
 
     /*
     /**********************************************************
-    /* Life-cycle
+    /* Bit of DoS detection goodness
+    /**********************************************************
+     */
+
+    /**
+     * Lazily constructed structure that is used to keep track of
+     * collision buckets that have overflowed once: this is used
+     * to detect likely attempts at denial-of-service attacks that
+     * uses hash collisions.
+     * 
+     * @since 2.4
+     */
+    private BitSet _overflows;
+
+    /*
+    /**********************************************************
+    /* Life-cycle: constructors
+    /**********************************************************
+     */
+
+    /**
+     * Main method for constructing a root symbol table instance.
+     */
+    private CharsToNameCanonicalizer(int seed)
+    {
+        _parent = null;
+        _seed = seed;
+        
+        // these settings don't really matter for the bootstrap instance
+        _canonicalize = true;
+        _flags = -1;
+        // And we'll also set flags so no copying of buckets is needed:
+        _hashShared = false; // doesn't really matter for root instance
+        _longestCollisionList = 0;
+
+        _tableInfo = new AtomicReference<TableInfo>(TableInfo.createInitial(DEFAULT_T_SIZE));
+        // and actually do NOT assign buffers so we'll find if anyone tried to
+        // use root instance
+    }
+
+    /**
+     * Internal constructor used when creating child instances.
+     */
+    private CharsToNameCanonicalizer(CharsToNameCanonicalizer parent, int flags, int seed,
+            TableInfo parentState)
+    {
+        _parent = parent;
+        _seed = seed;
+        _tableInfo = null; // not used by child tables
+        _flags = flags;
+        _canonicalize = JsonFactory.Feature.CANONICALIZE_FIELD_NAMES.enabledIn(flags);
+
+        // Then copy shared state
+        _symbols = parentState.symbols;
+        _buckets = parentState.buckets;
+
+        _size = parentState.size;
+        _longestCollisionList = parentState.longestCollisionList;
+
+        // Hard-coded fill factor, 75%
+        int arrayLen = (_symbols.length);
+        _sizeThreshold = _thresholdSize(arrayLen);
+        _indexMask =  (arrayLen - 1);
+
+        // Need to make copies of arrays, if/when adding new entries
+        _hashShared = true;
+    }
+
+    private static int _thresholdSize(int hashAreaSize) { return hashAreaSize - (hashAreaSize >> 2); }
+
+    /*
+    /**********************************************************
+    /* Life-cycle: factory methods, merging
     /**********************************************************
      */
 
@@ -209,81 +286,19 @@ public final class CharsToNameCanonicalizer
      * instance. Root instance is never used directly; its main use is for
      * storing and sharing underlying symbol arrays as needed.
      */
-    public static CharsToNameCanonicalizer createRoot()
-    {
-        /* [Issue-21]: Need to use a variable seed, to thwart hash-collision
-         * based attacks.
-         */
+    public static CharsToNameCanonicalizer createRoot() {
+        // Need to use a variable seed, to thwart hash-collision based attacks.
+        // 14-Feb-2017, tatu: not sure it actually helps, at all, since it won't
+        //   change mixing or any of the steps. Should likely just remove in future.
         long now = System.currentTimeMillis();
         // ensure it's not 0; and might as well require to be odd so:
         int seed = (((int) now) + ((int) (now >>> 32))) | 1;
         return createRoot(seed);
     }
-    
-    protected static CharsToNameCanonicalizer createRoot(int hashSeed) {
-        return sBootstrapSymbolTable.makeOrphan(hashSeed);
+
+    protected static CharsToNameCanonicalizer createRoot(int seed) {
+        return new CharsToNameCanonicalizer(seed);
     }
-
-    /**
-     * Main method for constructing a master symbol table instance.
-     *
-     * @param initialSize Minimum initial size for bucket array; internally
-     *   will always use a power of two equal to or bigger than this value.
-     */
-    private CharsToNameCanonicalizer()
-    {
-        // these settings don't really matter for the bootstrap instance
-        _canonicalize = true;
-        _intern = true;
-        // And we'll also set flags so no copying of buckets is needed:
-        _dirty = true;
-        _hashSeed = 0;
-        _longestCollisionList = 0;
-        initTables(DEFAULT_TABLE_SIZE);
-    }
-
-    private void initTables(int initialSize)
-    {
-        _symbols = new String[initialSize];
-        _buckets = new Bucket[initialSize >> 1];
-        // Mask is easy to calc for powers of two.
-        _indexMask = initialSize - 1;
-        _size = 0;
-        _longestCollisionList = 0;
-        // Hard-coded fill factor is 75%
-        _sizeThreshold = _thresholdSize(initialSize);
-    }
-
-    private static int _thresholdSize(int hashAreaSize) {
-        return hashAreaSize - (hashAreaSize >> 2);
-    }
-    
-    /**
-     * Internal constructor used when creating child instances.
-     */
-    private CharsToNameCanonicalizer(CharsToNameCanonicalizer parent,
-            boolean canonicalize, boolean intern,
-            String[] symbols, Bucket[] buckets, int size,
-            int hashSeed, int longestColl)
-    {
-        _parent = parent;
-        _canonicalize = canonicalize;
-        _intern = intern;
-
-        _symbols = symbols;
-        _buckets = buckets;
-        _size = size;
-        _hashSeed = hashSeed;
-        // Hard-coded fill factor, 75%
-        int arrayLen = (symbols.length);
-        _sizeThreshold = _thresholdSize(arrayLen);
-        _indexMask =  (arrayLen - 1);
-        _longestCollisionList = longestColl;
-
-        // Need to make copies of arrays, if/when adding new entries
-        _dirty = false;
-    }
-
     /**
      * "Factory" method; will create a new child instance of this symbol
      * table. It will be a copy-on-write instance, ie. it will only use
@@ -296,35 +311,27 @@ public final class CharsToNameCanonicalizer
      * on which only makeChild/mergeChild are called, but instance itself
      * is not used as a symbol table.
      */
-    public CharsToNameCanonicalizer makeChild(final boolean canonicalize,
-            final boolean intern)
-    {
-        /* 24-Jul-2012, tatu: Trying to reduce scope of synchronization, assuming
-         *   that synchronizing construction is the (potentially) expensive part,
-         *   and not so much short copy-the-variables thing.
-         */
-        final String[] symbols;
-        final Bucket[] buckets;
-        final int size;
-        final int hashSeed;
-        final int longestCollisionList;
-        
-        synchronized (this) {
-            symbols = _symbols;
-            buckets = _buckets;
-            size = _size;
-            hashSeed = _hashSeed;
-            longestCollisionList = _longestCollisionList;
-        }
-        
-        return new CharsToNameCanonicalizer(this, canonicalize, intern,
-                symbols, buckets, size, hashSeed, longestCollisionList);
+    public CharsToNameCanonicalizer makeChild(int flags) {
+        return new CharsToNameCanonicalizer(this, flags, _seed, _tableInfo.get());
     }
 
-    private CharsToNameCanonicalizer makeOrphan(int seed)
-    {
-        return new CharsToNameCanonicalizer(null, true, true,
-                _symbols, _buckets, _size, seed, _longestCollisionList);
+    /**
+     * Method called by the using code to indicate it is done with this instance.
+     * This lets instance merge accumulated changes into parent (if need be),
+     * safely and efficiently, and without calling code having to know about parent
+     * information.
+     */
+    public void release() {
+        // If nothing has been added, nothing to do
+        if (!maybeDirty()) { return; }
+
+        // we will try to merge if child table has new entries
+        if (_parent != null && _canonicalize) { // canonicalize set to false if max size was reached
+            _parent.mergeChild(new TableInfo(this));
+            // Let's also mark this instance as dirty, so that just in
+            // case release was too early, there's no corruption of possibly shared data.
+            _hashShared = true;
+        }
     }
 
     /**
@@ -334,60 +341,25 @@ public final class CharsToNameCanonicalizer
      * Note that caller has to make sure symbol table passed in is
      * really a child or sibling of this symbol table.
      */
-    private void mergeChild(CharsToNameCanonicalizer child)
+    private void mergeChild(TableInfo childState)
     {
-        /* One caveat: let's try to avoid problems with
-         * degenerate cases of documents with generated "random"
-         * names: for these, symbol tables would bloat indefinitely.
-         * One way to do this is to just purge tables if they grow
-         * too large, and that's what we'll do here.
-         */
-        if (child.size() > MAX_ENTRIES_FOR_REUSE
-                || child._longestCollisionList > MAX_COLL_CHAIN_FOR_REUSE) {
-            // Should there be a way to get notified about this event, to log it or such?
-            // (as it's somewhat abnormal thing to happen)
-            // At any rate, need to clean up the tables, then:
-            synchronized (this) {
-                initTables(DEFAULT_TABLE_SIZE);
-                // Dirty flag... well, let's just clear it. Shouldn't really matter for master tables
-                // (which this is, given something is merged to it)
-                _dirty = false;
-            }
-        } else {
-            // Otherwise, we'll merge changed stuff in, if there are  more entries (which
-            // may not be the case if one of siblings has added symbols first or such)
-            if (child.size() <= size()) { // nothing to add
-                return;
-            }
-            // Okie dokie, let's get the data in!
-            synchronized (this) {
-                _symbols = child._symbols;
-                _buckets = child._buckets;
-                _size = child._size;
-                _sizeThreshold = child._sizeThreshold;
-                _indexMask = child._indexMask;
-                _longestCollisionList = child._longestCollisionList;
-                // Dirty flag... well, let's just clear it. Shouldn't really matter for master tables
-                // (which this is, given something is merged to it)
-                _dirty = false;
-            }
-        }
-    }
+        final int childCount = childState.size;
+        TableInfo currState = _tableInfo.get();
 
-    public void release()
-    {
-        // If nothing has been added, nothing to do
-        if (!maybeDirty()) {
+        // Should usually grow; but occasionally could also shrink if (but only if)
+        // collision list overflow ends up clearing some collision lists.
+        if (childCount == currState.size) {
             return;
         }
-        if (_parent != null) {
-            _parent.mergeChild(this);
-            /* Let's also mark this instance as dirty, so that just in
-             * case release was too early, there's no corruption
-             * of possibly shared data.
-             */
-            _dirty = false;
+        // One caveat: let's try to avoid problems with  degenerate cases of documents with
+        // generated "random" names: for these, symbol tables would bloat indefinitely.
+        // One way to do this is to just purge tables if they grow
+        // too large, and that's what we'll do here.
+        if (childCount > MAX_ENTRIES_FOR_REUSE) {
+            // At any rate, need to clean up the tables
+            childState = TableInfo.createInitial(DEFAULT_T_SIZE);
         }
+        _tableInfo.compareAndSet(currState, childState);
     }
 
     /*
@@ -396,7 +368,13 @@ public final class CharsToNameCanonicalizer
     /**********************************************************
      */
 
-    public int size() { return _size; }
+    public int size() {
+        if (_tableInfo != null) { // root table
+            return _tableInfo.get().size;
+        }
+        // nope, child table
+        return _size;
+    }
 
     /**
      * Method for checking number of primary hash buckets this symbol
@@ -404,13 +382,10 @@ public final class CharsToNameCanonicalizer
      * 
      * @since 2.1
      */
-    public int bucketCount() { 
-       return _symbols.length; }
-    
-    public boolean maybeDirty() { return _dirty; }
+    public int bucketCount() {  return _symbols.length; }
+    public boolean maybeDirty() { return !_hashShared; }
+    public int hashSeed() { return _seed; }
 
-    public int hashSeed() { return _hashSeed; }
-    
     /**
      * Method mostly needed by unit tests; calculates number of
      * entries that are in collision list. Value can be at most
@@ -418,13 +393,12 @@ public final class CharsToNameCanonicalizer
      * 
      * @since 2.1
      */
-    public int collisionCount()
-    {
+    public int collisionCount() {
         int count = 0;
         
         for (Bucket bucket : _buckets) {
             if (bucket != null) {
-                count += bucket.length();
+                count += bucket.length;
             }
         }
         return count;
@@ -437,10 +411,7 @@ public final class CharsToNameCanonicalizer
      * 
      * @since 2.1
      */
-    public int maxCollisionLength()
-    {
-        return _longestCollisionList;
-    }
+    public int maxCollisionLength() { return _longestCollisionList; }
 
     /*
     /**********************************************************
@@ -470,39 +441,54 @@ public final class CharsToNameCanonicalizer
             // Let's inline primary String equality checking:
             if (sym.length() == len) {
                 int i = 0;
-                do {
-                    if (sym.charAt(i) != buffer[start+i]) {
-                        break;
+                while (sym.charAt(i) == buffer[start+i]) {
+                    // Optimal case; primary match found
+                    if (++i == len) {
+                        return sym;
                     }
-                } while (++i < len);
-                // Optimal case; primary match found
-                if (i == len) {
-                    return sym;
                 }
             }
-            // How about collision bucket?
-            Bucket b = _buckets[index >> 1];
+            Bucket b = _buckets[index>>1];
             if (b != null) {
-                sym = b.find(buffer, start, len);
+                sym = b.has(buffer, start, len);
+                if (sym != null) {
+                    return sym;
+                }
+                sym = _findSymbol2(buffer, start, len, b.next);
                 if (sym != null) {
                     return sym;
                 }
             }
         }
+        return _addSymbol(buffer, start, len, h, index);
+    }
 
-        if (!_dirty) { //need to do copy-on-write?
+    private String _findSymbol2(char[] buffer, int start, int len, Bucket b) {
+        while (b != null) {
+            String sym = b.has(buffer, start, len);
+            if (sym != null) {
+                return sym;
+            }
+            b = b.next;
+        }
+        return null;
+    }
+
+    private String _addSymbol(char[] buffer, int start, int len, int h, int index)
+    {
+        if (_hashShared) { //need to do copy-on-write?
             copyArrays();
-            _dirty = true;
+            _hashShared = false;
         } else if (_size >= _sizeThreshold) { // Need to expand?
-           rehash();
-           /* Need to recalc hash; rare occurence (index mask has been
-            * recalculated as part of rehash)
-            */
-           index = _hashToIndex(calcHash(buffer, start, len));
+            rehash();
+            /* Need to recalc hash; rare occurence (index mask has been
+             * recalculated as part of rehash)
+             */
+            index = _hashToIndex(calcHash(buffer, start, len));
         }
 
         String newSymbol = new String(buffer, start, len);
-        if (_intern) {
+        if (JsonFactory.Feature.INTERN_FIELD_NAMES.enabledIn(_flags)) {
             newSymbol = InternCache.instance.intern(newSymbol);
         }
         ++_size;
@@ -510,28 +496,59 @@ public final class CharsToNameCanonicalizer
         if (_symbols[index] == null) {
             _symbols[index] = newSymbol;
         } else {
-            int bix = (index >> 1);
+            final int bix = (index >> 1);
             Bucket newB = new Bucket(newSymbol, _buckets[bix]);
-            _buckets[bix] = newB;
-            _longestCollisionList = Math.max(newB.length(), _longestCollisionList);
-            if (_longestCollisionList > MAX_COLL_CHAIN_LENGTH) {
-                reportTooManyCollisions(MAX_COLL_CHAIN_LENGTH);
+            int collLen = newB.length;
+            if (collLen > MAX_COLL_CHAIN_LENGTH) {
+                // 23-May-2014, tatu: Instead of throwing an exception right away,
+                //    let's handle in bit smarter way.
+                _handleSpillOverflow(bix, newB);
+            } else {
+                _buckets[bix] = newB;
+                _longestCollisionList = Math.max(collLen, _longestCollisionList);
             }
         }
-
         return newSymbol;
     }
 
+    private void _handleSpillOverflow(int bindex, Bucket newBucket)
+    {
+        if (_overflows == null) {
+            _overflows = new BitSet();
+            _overflows.set(bindex);
+        } else {
+            if (_overflows.get(bindex)) {
+                // Has happened once already, so not a coincident...
+                if (JsonFactory.Feature.FAIL_ON_SYMBOL_HASH_OVERFLOW.enabledIn(_flags)) {
+                    reportTooManyCollisions(MAX_COLL_CHAIN_LENGTH);
+                }
+                // but even if we don't fail, we will stop canonicalizing:
+                _canonicalize = false;
+            } else {
+                _overflows.set(bindex);
+            }
+        }
+        // regardless, if we get this far, clear up the bucket, adjust size appropriately.
+        _symbols[bindex + bindex] = newBucket.symbol;
+        _buckets[bindex] = null;
+        // newBucket contains new symbol; but we wil 
+        _size -= (newBucket.length);
+        // we could calculate longest; but for now just mark as invalid
+        _longestCollisionList = -1;
+    }
+
     /**
      * Helper method that takes in a "raw" hash value, shuffles it as necessary,
      * and truncates to be used as the index.
      */
-    public int _hashToIndex(int rawHash)
-    {
-        rawHash += (rawHash >>> 15); // this seems to help quite a bit, at least for our tests
+    public int _hashToIndex(int rawHash) {
+        // doing these seems to help a bit
+        rawHash += (rawHash >>> 15);
+        rawHash ^= (rawHash << 7);
+        rawHash += (rawHash >>> 3);
         return (rawHash & _indexMask);
     }
-    
+
     /**
      * Implementation of a hashing method for variable length
      * Strings. Most of the time intention is that this calculation
@@ -541,10 +558,9 @@ public final class CharsToNameCanonicalizer
      * @param len Length of String; has to be at least 1 (caller guarantees
      *   this pre-condition)
      */
-    public int calcHash(char[] buffer, int start, int len)
-    {
-        int hash = _hashSeed;
-        for (int i = 0; i < len; ++i) {
+    public int calcHash(char[] buffer, int start, int len) {
+        int hash = _seed;
+        for (int i = start, end = start+len; i < end; ++i) {
             hash = (hash * HASH_MULT) + (int) buffer[i];
         }
         // NOTE: shuffling, if any, is done in 'findSymbol()', not here:
@@ -555,7 +571,7 @@ public final class CharsToNameCanonicalizer
     {
         final int len = key.length();
         
-        int hash = _hashSeed;
+        int hash = _seed;
         for (int i = 0; i < len; ++i) {
             hash = (hash * HASH_MULT) + (int) key.charAt(i);
         }
@@ -574,14 +590,10 @@ public final class CharsToNameCanonicalizer
      * change is made to a derived symbol table.
      */
     private void copyArrays() {
-        String[] oldSyms = _symbols;
-        int size = oldSyms.length;
-        _symbols = new String[size];
-        System.arraycopy(oldSyms, 0, _symbols, 0, size);
-        Bucket[] oldBuckets = _buckets;
-        size = oldBuckets.length;
-        _buckets = new Bucket[size];
-        System.arraycopy(oldBuckets, 0, _buckets, 0, size);
+        final String[] oldSyms = _symbols;
+        _symbols = Arrays.copyOf(oldSyms, oldSyms.length);
+        final Bucket[] oldBuckets = _buckets;
+        _buckets = Arrays.copyOf(oldBuckets, oldBuckets.length);
     }
 
     /**
@@ -591,8 +603,7 @@ public final class CharsToNameCanonicalizer
      * is really redistributing old entries into new String/Bucket
      * entries.
      */
-    private void rehash()
-    {
+    private void rehash() {
         int size = _symbols.length;
         int newSize = size + size;
 
@@ -600,18 +611,19 @@ public final class CharsToNameCanonicalizer
          *    prepared to use, to guard against OOME in case of unbounded
          *    name sets (unique [non-repeating] names)
          */
-        if (newSize > MAX_TABLE_SIZE) {
-            /* If this happens, there's no point in either growing or
-             * shrinking hash areas. Rather, it's better to just clean
-             * them up for reuse.
-             */
+        if (newSize > MAX_T_SIZE) {
+            // If this happens, there's no point in either growing or shrinking hash areas.
+            // Rather, let's just cut our losses and stop canonicalizing.
             _size = 0;
-            Arrays.fill(_symbols, null);
-            Arrays.fill(_buckets, null);
-            _dirty = true;
+            _canonicalize = false;
+            // in theory, could just leave these as null, but...
+            _symbols = new String[DEFAULT_T_SIZE];
+            _buckets = new Bucket[DEFAULT_T_SIZE>>1];
+            _indexMask = DEFAULT_T_SIZE-1;
+            _hashShared = false;
             return;
         }
-        
+
         String[] oldSyms = _symbols;
         Bucket[] oldBuckets = _buckets;
         _symbols = new String[newSize];
@@ -619,12 +631,11 @@ public final class CharsToNameCanonicalizer
         // Let's update index mask, threshold, now (needed for rehashing)
         _indexMask = newSize - 1;
         _sizeThreshold = _thresholdSize(newSize);
-        
+
         int count = 0; // let's do sanity check
 
-        /* Need to do two loops, unfortunately, since spill-over area is
-         * only half the size:
-         */
+        // Need to do two loops, unfortunately, since spill-over area is
+        // only half the size:
         int maxColl = 0;
         for (int i = 0; i < size; ++i) {
             String symbol = oldSyms[i];
@@ -637,7 +648,7 @@ public final class CharsToNameCanonicalizer
                     int bix = (index >> 1);
                     Bucket newB = new Bucket(symbol, _buckets[bix]);
                     _buckets[bix] = newB;
-                    maxColl = Math.max(maxColl, newB.length());
+                    maxColl = Math.max(maxColl, newB.length);
                 }
             }
         }
@@ -647,7 +658,7 @@ public final class CharsToNameCanonicalizer
             Bucket b = oldBuckets[i];
             while (b != null) {
                 ++count;
-                String symbol = b.getSymbol();
+                String symbol = b.symbol;
                 int index = _hashToIndex(calcHash(symbol));
                 if (_symbols[index] == null) {
                     _symbols[index] = symbol;
@@ -655,30 +666,81 @@ public final class CharsToNameCanonicalizer
                     int bix = (index >> 1);
                     Bucket newB = new Bucket(symbol, _buckets[bix]);
                     _buckets[bix] = newB;
-                    maxColl = Math.max(maxColl, newB.length());
+                    maxColl = Math.max(maxColl, newB.length);
                 }
-                b = b.getNext();
+                b = b.next;
             }
         }
         _longestCollisionList = maxColl;
+        _overflows = null;
 
         if (count != _size) {
-            throw new Error("Internal error on SymbolTable.rehash(): had "+_size+" entries; now have "+count+".");
+            throw new IllegalStateException(String.format(
+                    "Internal error on SymbolTable.rehash(): had %d entries; now have %d",
+                    _size, count));
         }
     }
 
     /**
      * @since 2.1
      */
-    protected void reportTooManyCollisions(int maxLen)
-    {
+    protected void reportTooManyCollisions(int maxLen) {
         throw new IllegalStateException("Longest collision chain in symbol table (of size "+_size
                 +") now exceeds maximum, "+maxLen+" -- suspect a DoS attack based on hash collisions");
     }
-    
+
+    // For debugging, comment out
+    /*
+    @Override
+    public String toString()
+    {
+        StringBuilder sb = new StringBuilder();
+        int primaryCount = 0;
+        for (String s : _symbols) {
+            if (s != null) ++primaryCount;
+        }
+        
+        sb.append("[BytesToNameCanonicalizer, size: ");
+        sb.append(_size);
+        sb.append('/');
+        sb.append(_symbols.length);
+        sb.append(", ");
+        sb.append(primaryCount);
+        sb.append('/');
+        sb.append(_size - primaryCount);
+        sb.append(" coll; avg length: ");
+
+        // Average length: minimum of 1 for all (1 == primary hit);
+        // and then 1 per each traversal for collisions/buckets
+        //int maxDist = 1;
+        int pathCount = _size;
+        for (Bucket b : _buckets) {
+            if (b != null) {
+                int spillLen = b.length;
+                for (int j = 1; j <= spillLen; ++j) {
+                    pathCount += j;
+                }
+            }
+        }
+        double avgLength;
+
+        if (_size == 0) {
+            avgLength = 0.0;
+        } else {
+            avgLength = (double) pathCount / (double) _size;
+        }
+        // let's round up a bit (two 2 decimal places)
+        //avgLength -= (avgLength % 0.01);
+
+        sb.append(avgLength);
+        sb.append(']');
+        return sb.toString();
+    }
+*/
+
     /*
     /**********************************************************
-    /* Bucket class
+    /* Helper classes
     /**********************************************************
      */
 
@@ -688,43 +750,65 @@ public final class CharsToNameCanonicalizer
      */
     static final class Bucket
     {
-        private final String _symbol;
-        private final Bucket _next;
-        private final int _length;
+        public final String symbol;
+        public final Bucket next;
+        public final int length;
 
-        public Bucket(String symbol, Bucket next) {
-            _symbol = symbol;
-            _next = next;
-            _length = (next == null) ? 1 : next._length+1;
+        public Bucket(String s, Bucket n) {
+            symbol = s;
+            next = n;
+            length = (n == null) ? 1 : n.length+1;
         }
 
-        public String getSymbol() { return _symbol; }
-        public Bucket getNext() { return _next; }
-        public int length() { return _length; }
-
-        public String find(char[] buf, int start, int len) {
-            String sym = _symbol;
-            Bucket b = _next;
-
-            while (true) { // Inlined equality comparison:
-                if (sym.length() == len) {
-                    int i = 0;
-                    do {
-                        if (sym.charAt(i) != buf[start+i]) {
-                            break;
-                        }
-                    } while (++i < len);
-                    if (i == len) {
-                        return sym;
-                    }
-                }
-                if (b == null) {
-                    break;
-                }
-                sym = b.getSymbol();
-                b = b.getNext();
+        public String has(char[] buf, int start, int len) {
+            if (symbol.length() != len) {
+                return null;
             }
-            return null;
+            int i = 0;
+            do {
+                if (symbol.charAt(i) != buf[start+i]) {
+                    return null;
+                }
+            } while (++i < len);
+            return symbol;
+        }
+    }
+
+    /**
+     * Immutable value class used for sharing information as efficiently
+     * as possible, by only require synchronization of reference manipulation
+     * but not access to contents.
+     * 
+     * @since 2.8.7
+     */
+    private final static class TableInfo
+    {
+        final int size;
+        final int longestCollisionList;
+        final String[] symbols;
+        final Bucket[] buckets;
+
+        public TableInfo(int size, int longestCollisionList,
+                String[] symbols, Bucket[] buckets)
+        {
+            this.size = size;
+            this.longestCollisionList = longestCollisionList;
+            this.symbols = symbols;
+            this.buckets = buckets;
+        }
+
+        public TableInfo(CharsToNameCanonicalizer src)
+        {
+            this.size = src._size;
+            this.longestCollisionList = src._longestCollisionList;
+            this.symbols = src._symbols;
+            this.buckets = src._buckets;
+        }
+
+        public static TableInfo createInitial(int sz) {
+            return new TableInfo(0,
+                    0, // longestCollisionList
+                    new String[sz], new Bucket[sz >> 1]);
         }
     }
 }
